# -*- coding: utf-8 -*-
import argparse
import glob
import logging
import os
import re
import socket
import sys

from requires_io import __version__, consts
from requires_io.api import RequiresAPI
from requires_io.draw import draw

log = logging.getLogger(__name__)


class NameType(object):
    def __init__(self):
        self.max_length = 128
        self.pattern = '^[a-zA-Z0-9-_\.]{1,%s}$' % self.max_length

    def __call__(self, value):
        if not value:
            raise argparse.ArgumentTypeError('required')
        if len(value) > self.max_length:
            raise argparse.ArgumentTypeError('at most %d character (it has %d)' % (self.max_length, len(value)))
        if not re.search(self.pattern, value):
            raise argparse.ArgumentTypeError('only alphanumeric and "-_." character allowed')
        return value


class TokenType(object):
    def __call__(self, value):
        if not value:
            raise argparse.ArgumentTypeError('required')
        return value


glob_type_re = re.compile(r'''
   [/\\](
   setup\.py
   |tox\.ini
   |(buildout|versions)\.cfg
   |(req|pip|dep)[\w-]*\.(txt|pip)
   |requirements[/\\][\w-]*\.(pip|txt)
   )$
''', re.X | re.IGNORECASE)


class GlobType(object):
    def __call__(self, value):
        paths = {}
        for path in glob.glob(os.path.normpath(os.path.abspath(value))):
            # If this is a folder, look for known files in it
            if os.path.isdir(path):
                for root, dirs, files in os.walk(path):
                    # Go threw the files to check if they match a known pattern
                    for name in files:
                        self._store(paths, os.path.join(root, name))
                    # Ignore CVS folders
                    if 'CVS' in dirs:
                        dirs.remove('CVS')
                    # Ignore hidden folders
                    ignores = [d for d in dirs if d.startswith('.')]
                    for ignore in ignores:
                        dirs.remove(ignore)
            # If this is a file, add it if matches the pattern
            elif os.path.isfile(path):
                self._store(paths, path)
        if not paths:
            raise argparse.ArgumentTypeError('failed to find requirement files matching the pattern for %s' % value)
        return paths

    def _store(self, paths, path):
        match = glob_type_re.search(path)
        if match:
            paths[path] = match.group(1)


class PathType(object):
    def __call__(self, value):
        if not os.path.isfile(value):
            raise argparse.ArgumentTypeError('file does not exist: %s' % value)
        return value


def _common_index(paths):
    index = 0
    for parts in zip(*[os.path.normcase(path).split(os.sep) for path in paths]):
        if len(set(parts)) != 1:
            return index
        index += 1
    return index


def _to_urls(*all_paths):
    paths = {}
    for currents_paths in all_paths:
        paths.update(currents_paths)
    if not paths:
        return {}
    candidates = set()
    for path, basename in paths.items():
        candidates.add(path[:-len(basename) - 1])
    candidates = list(candidates)
    index = _common_index(candidates)
    urls = {}
    for path in paths.keys():
        url = '/'.join(os.path.splitdrive(path)[1].split(os.sep)[index:])
        urls[path] = url
    return urls


class Commands(object):
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog='requires.io')
        self.parser.add_argument('--version', action='version', version=__version__)
        self.subparsers = self.parser.add_subparsers()
        self.add_parser_update_repository()
        self.add_parser_delete_repository()
        self.add_parser_update_branch()
        self.add_parser_delete_branch()
        self.add_parser_update_tag()
        self.add_parser_delete_tag()
        self.add_parser_update_site()
        self.add_parser_delete_site()
        self.add_parser_parse()

    def add_argument_paths(self, group):
        group.add_argument('paths', metavar='PATH', type=GlobType(), nargs='+',
                           help='requirement files or folders containing requirement files (glob allowed)')

    def add_parser(self, title, help, executor):
        parser = self.subparsers.add_parser(title, help=help)
        group = parser.add_argument_group('global options')
        group.add_argument('-t', '--token',
                           help='API token (default: REQUIRES_TOKEN environment variable)',
                           type=TokenType(), default=os.getenv('REQUIRES_TOKEN'))
        parser.set_defaults(execute=lambda args: executor(RequiresAPI(args.token), args))
        return parser.add_argument_group('command options')

    def add_repository_parser(self, *args, **kwargs):
        group = self.add_parser(*args, **kwargs)
        group.add_argument('-r', '--repository', metavar='REPO',
                           help='repository name',
                           type=NameType(), required=True)
        return group

    def execute(self, args):
        args = self.parser.parse_args(args)
        if not hasattr(args, 'execute'):
            self.parser.print_usage()
        else:
            args.execute(args)

    # =========================================================================
    # REPOSITORY
    # -------------------------------------------------------------------------
    def add_parser_update_repository(self):
        group = self.add_repository_parser('update-repo', 'update repository',
                                           lambda api, args: api.update_repository(args.repository, args.private))
        subgroup = group.add_mutually_exclusive_group(required=True)
        subgroup.add_argument('--public', dest='private', help='repo public', action='store_false')
        subgroup.add_argument('--private', dest='private', help='repo public', action='store_true')

    def add_parser_delete_repository(self):
        self.add_repository_parser('delete-repo', 'delete repository',
                                   lambda api, args: api.delete_repository(args.repository))

    # =========================================================================
    # BRANCH
    # -------------------------------------------------------------------------
    def add_argument_branch_name(self, group):
        group.add_argument('-n', '--name', help='branch name', type=NameType(), required=True)

    def add_parser_update_branch(self):
        group = self.add_repository_parser('update-branch', 'create or update branch',
                                           lambda api, args: api.update_branch(args.repository, args.name,
                                                                               _to_urls(*args.paths)))
        self.add_argument_branch_name(group)
        self.add_argument_paths(group)

    def add_parser_delete_branch(self):
        group = self.add_repository_parser('delete-branch', 'delete branch',
                                           lambda api, args: api.delete_branch(args.repository, args.name))
        self.add_argument_branch_name(group)

    # =========================================================================
    # TAG
    # -------------------------------------------------------------------------
    def add_argument_tag_name(self, group):
        group.add_argument('-n', '--name', help='tag name', type=NameType(), required=True)

    def add_parser_update_tag(self):
        group = self.add_repository_parser('update-tag', 'create or update tag',
                                           lambda api, args: api.update_tag(args.repository, args.name,
                                                                            _to_urls(*args.paths)))
        self.add_argument_tag_name(group)
        self.add_argument_paths(group)

    def add_parser_delete_tag(self):
        group = self.add_repository_parser('delete-tag', 'delete tag',
                                           lambda api, args: api.delete_tag(args.repository, args.name))
        self.add_argument_tag_name(group)

    # =========================================================================
    # SITE
    # -------------------------------------------------------------------------
    def add_argument_site_name(self, group):
        group.add_argument('-n', '--name', help='site name (default: %s)' % socket.gethostname(),
                           type=NameType(), default=socket.gethostname())

    def add_parser_update_site(self):
        group = self.add_repository_parser('update-site', 'create or update site',
                                           lambda api, args: api.update_site(args.repository, args.name))
        self.add_argument_site_name(group)

    def add_parser_delete_site(self):
        group = self.add_repository_parser('delete-site', 'delete site',
                                           lambda api, args: api.delete_site(args.repository, args.name))
        self.add_argument_site_name(group)

    # =========================================================================
    # REQUIREMENTS
    # -------------------------------------------------------------------------
    def add_parser_parse(self):
        group = self.add_parser('parse', 'parse provided requirements file',
                                lambda api, args: draw(api.get_requirements(args.paths[0], args.file_type)))
        group.add_argument('-k', '--kind', dest='file_type', choices=consts.TYPES,
                           help='type of requirements file (default: let us guess from file name)')
        group.add_argument('paths', metavar='PATH', type=PathType(), nargs=1,
                           help='requirements file to process')


def main(args=sys.argv, setup_log=True):
    if setup_log:
        logging.basicConfig(format='%(message)s', level=logging.INFO)
        logging.getLogger('requests').setLevel(logging.WARNING)
    Commands().execute(args[1:])



# -*- coding: utf-8 -*-
import sys
import json
import base64
import codecs
import logging
import subprocess

import requests

from requires_io import consts


log = logging.getLogger(__name__)

try:
    check_output = subprocess.check_output
except AttributeError:
    def check_output(*args, **kwargs):
        process = subprocess.Popen(stdout=subprocess.PIPE, *args, **kwargs)
        output, unused_err = process.communicate()
        code = process.poll()
        if code:
            cmd = kwargs.get('args')
            if cmd is None:
                cmd = args[0]
            raise subprocess.CalledProcessError(code, cmd, output=output)
        return output


class RequiresAPI(object):
    def __init__(self, token, base_url='https://requires.io/api/v2/', verify=True):
        self.token = token
        self.base_url = base_url
        if self.base_url[-1] != '/':
            self.base_url += '/'
        self.verify = verify

    def _get_headers(self, content_type='application/json'):
        headers = {
            'Authorization': 'Token %s' % self.token,
            'Accept': 'application/json',
        }
        if content_type:
            headers['Content-Type'] = content_type
        return headers

    def _update_reference(self, url, paths):
        payload = []
        for path, relative in paths.items():
            log.info('add %s to payload', relative)
            with open(path, 'rb') as f:
                payload.append({
                    'path': relative,
                    'content': base64.b64encode(f.read()).decode('utf-8'),
                })
        requests.put(
            url,
            headers=self._get_headers(),
            data=json.dumps(payload),
            verify=self.verify,
        ).raise_for_status()

    # =========================================================================
    # REPOSITORY
    # -------------------------------------------------------------------------
    def _get_repository_url(self, repository):
        return '%srepos/%s' % (self.base_url, repository)

    def update_repository(self, repository, private):
        payload = dict(
            private=private,
        )
        requests.put(
            self._get_repository_url(repository),
            headers=self._get_headers(),
            data=json.dumps(payload),
            verify=self.verify,
        ).raise_for_status()

    def delete_repository(self, repository):
        log.info('delete repository %s', repository)
        requests.delete(
            self._get_repository_url(repository),
            headers=self._get_headers(),
            verify=self.verify,
        ).raise_for_status()

    # =========================================================================
    # BRANCH
    # -------------------------------------------------------------------------
    def _get_branch_url(self, repository, name):
        return '%s/branches/%s' % (self._get_repository_url(repository), name)

    def update_branch(self, repository, name, paths):
        log.info('update branch %s on repository %s', name, repository)
        return self._update_reference(
            self._get_branch_url(repository, name),
            paths,
        )

    def delete_branch(self, repository, name):
        log.info('delete branch %s on repository %s', name, repository)
        requests.delete(
            self._get_branch_url(repository, name),
            headers=self._get_headers(),
            verify=self.verify,
        ).raise_for_status()

    # =========================================================================
    # TAG
    # -------------------------------------------------------------------------
    def _get_tag_url(self, repository, name):
        return '%s/tags/%s' % (self._get_repository_url(repository), name)

    def update_tag(self, repository, name, paths):
        log.info('update tag %s on repository %s', name, repository)
        return self._update_reference(
            self._get_tag_url(repository, name),
            paths,
        )

    def delete_tag(self, repository, name):
        log.info('delete tag %s on repository %s', name, repository)
        requests.delete(
            self._get_tag_url(repository, name),
            headers=self._get_headers(),
            verify=self.verify,
        ).raise_for_status()

    # =========================================================================
    # SITE
    # -------------------------------------------------------------------------
    def _get_site_url(self, repository, name):
        return '%s/sites/%s' % (self._get_repository_url(repository), name)

    def update_site(self, repository, name):
        log.info('update site %s on repository %s', name, repository)
        command = ['pip', 'freeze', '--local']
        output = check_output(command)
        encoding = getattr(sys.stdout, 'encoding', 'utf-8')
        data = codecs.decode(output, encoding, 'replace')
        requests.put(
            self._get_site_url(repository, name),
            headers=self._get_headers('text/plain'),
            data=data,
            verify=self.verify,
        ).raise_for_status()

    def delete_site(self, repository, name):
        log.info('delete site %s on repository %s', name, repository)
        requests.delete(
            self._get_site_url(repository, name),
            headers=self._get_headers(),
            verify=self.verify,
        ).raise_for_status()

    # =========================================================================
    # REQUIREMENTS
    # -------------------------------------------------------------------------
    def get_requirements(self, file_path, file_type=None):
        data = {}
        if file_type:
            if file_type not in consts.TYPES:
                raise ValueError('invalid file type: %s' % file_type)
            data['file_type'] = file_type
        with open(file_path, 'rb') as fd:
            response = requests.post(
                self.base_url + 'requirements/',
                files={'file': fd},
                data=data,
                headers=self._get_headers(content_type=None),
                verify=self.verify,
            )
            response.raise_for_status()
            return response.json()



# -*- coding: utf-8 -*-
import sys


def column_length(requirements, title, key):
    l = lambda r: len(key(r))
    if not requirements:
        return len(title)
    return max(len(title), l(max(requirements, key=l)))


def draw(requirements, stream=sys.stdout):
    t_name = 'Package'
    t_specs = 'Requirement'
    t_latest_version = 'Latest'
    t_status = 'Status'
    l_name = column_length(requirements, t_name, lambda r: r['package']['name'])
    l_specs = column_length(requirements, t_specs, lambda r: r['specs'])
    l_latest_version = column_length(requirements, t_latest_version, lambda r: r['latest']['version'])
    l_status = column_length(requirements, t_status, lambda r: r['status'])
    string_format = '%%-%ds  %%-%ds  %%-%ds  %%-%ds\n' % (l_name, l_specs, l_latest_version, l_status)
    separator = string_format % (l_name * '=', l_specs * '=', l_latest_version * '=', l_status * '=')
    stream.write(separator)
    stream.write(string_format % (t_name, t_specs, t_latest_version, t_status))
    stream.write(separator)
    if requirements:
        for r in requirements:
            stream.write(string_format % (r['package']['name'], r['specs'], r['latest']['version'], r['status']))
    else:
        stream.write(string_format % ('-', '-', '-', '-'))
    stream.write(separator)



# -*- coding: utf-8 -*-
import codecs
import contextlib
import os
import shutil
import tempfile
import unittest

from requests.exceptions import HTTPError
from requests.status_codes import codes

from requires_io.commands import glob_type_re, GlobType, main, _to_urls


class Repository(object):
    def __init__(self, name):
        self.name = name
        self.root = None

    @contextlib.contextmanager
    def context(self):
        tmp = tempfile.mkdtemp()
        try:
            self.root = os.path.join(tmp, self.name)
            yield self
        finally:
            shutil.rmtree(tmp)
            self.root = None

    def write(self, filename, content, **kwargs):
        path = os.path.normpath(os.path.join(self.root, filename))
        fld = os.path.dirname(path)
        if not os.path.exists(fld):
            os.makedirs(fld)
        with codecs.open(path, kwargs.pop('mode', 'w'), **kwargs) as fd:
            fd.write(content)
            if kwargs.get('clrf', True):
                if not content.endswith('\n'):
                    fd.write('\n')


class TestCase(unittest.TestCase):
    def assertIsNotNone(self, val):  # missing in 2.6
        self.assertTrue(val is not None)

    def assertRaiseForStatus(self, status_code, callable_obj, *args):
        try:
            callable_obj(*args)
        except HTTPError as e:
            self.assertEquals(status_code, e.response.status_code)
        else:
            self.fail('failed to raise status code %d' % status_code)

    def test_re(self):
        self.assertIsNotNone(glob_type_re.search('/foo/bar/setup.py'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/tox.ini'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/buildout.cfg'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/versions.cfg'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/requirements.txt'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/requirements.pip'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/pip.txt'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/dependences.txt'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/requirements/prod.txt'))
        self.assertIsNotNone(glob_type_re.search('/foo/bar/requirements/test.pip'))

    def test_to_url(self):
        n = os.path.normpath
        self.assertEquals({}, _to_urls({}))
        self.assertEquals({
            n('/root/foo/bar/setup.py'): 'setup.py',
        }, _to_urls({
            n('/root/foo/bar/setup.py'): 'setup.py',
        }))
        self.assertEquals({
            n('/root/foo/bar/setup.py'): 'setup.py',
            n('/root/foo/bar/requirements/prod.txt'): 'requirements/prod.txt',
        }, _to_urls({
            n('/root/foo/bar/setup.py'): 'setup.py',
            n('/root/foo/bar/requirements/prod.txt'): n('requirements/prod.txt'),
        }))
        self.assertEquals({
            n('/root/foo/bar/requirements/dev.txt'): 'requirements/dev.txt',
            n('/root/foo/bar/requirements/prod.txt'): 'requirements/prod.txt',
        }, _to_urls({
            n('/root/foo/bar/requirements/dev.txt'): n('requirements/dev.txt'),
            n('/root/foo/bar/requirements/prod.txt'): n('requirements/prod.txt'),
        }))
        self.assertEquals({
            n('/root/foo/baz/setup.py'): 'baz/setup.py',
            n('/root/foo/bar/requirements/prod.txt'): 'bar/requirements/prod.txt',
        }, _to_urls({
            n('/root/foo/baz/setup.py'): n('setup.py'),
            n('/root/foo/bar/requirements/prod.txt'): n('requirements/prod.txt'),
        }))
        self.assertEquals({
            n('/root/foo/baz/setup.py'): 'baz/setup.py',
            n('/root/foo/bar/requirements/prod.txt'): 'bar/requirements/prod.txt',
        }, _to_urls({
            n('/root/foo/baz/setup.py'): n('setup.py'),
            n('/root/foo/bar/requirements/prod.txt'): n('requirements/prod.txt'),
        }))

    def assertPaths(self, paths, path):
        self.assertEquals(paths, GlobType()(path))

    def test_paths(self):
        j = os.path.join
        repository = Repository('foo')
        with repository.context():
            repository.write('setup.py', 'hello')
            repository.write(j('requirements', 'prod.txt'), 'hello')
            self.assertPaths({
                j(repository.root, 'setup.py'): 'setup.py',
                j(repository.root, 'requirements', 'prod.txt'): j('requirements', 'prod.txt'),
            }, repository.root)
            self.assertPaths({
                j(repository.root, 'requirements', 'prod.txt'): j('requirements', 'prod.txt'),
            }, j(repository.root, '*', '*.txt'))

    def test_update_site(self):
        self.assertRaiseForStatus(codes.UNAUTHORIZED, main, ['requires.io', 'update-site', '-t', '1234', '-r', 'foo'])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import os
import codecs

# allow setup.py to be run from any path
os.chdir(os.path.dirname(os.path.abspath(__file__)))

try:
    import setuptools
except ImportError:
    import ez_setup

    ez_setup.use_setuptools()
from setuptools import setup

install_requires = ['requests >= 2.0.0', ]

try:
    import argparse
except ImportError:
    install_requires.append('argparse >= 1.2.0')

with codecs.open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()
with codecs.open('CHANGES.rst', 'r', 'utf-8') as f:
    history = f.read()

setup(
    name='requires.io',
    version='0.2.6',
    description='Requires.io API',
    long_description=readme + '\n\n' + history,
    url='https://requires.io/',
    author='Requires.io',
    author_email='support@requires.io',
    packages=['requires_io', ],
    include_package_data=False,
    zip_safe=True,
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'requires.io = requires_io.commands:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)

